# Copyright (C) 2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import AsyncIterator, Iterable, Optional

from pontos.github.api.client import GitHubAsyncREST
from pontos.github.models.base import (
    Permission,
    Team,
    TeamPrivacy,
    TeamRole,
    User,
)
from pontos.github.models.organization import Repository


class GitHubAsyncRESTTeams(GitHubAsyncREST):
    async def get_all(self, organization: str) -> AsyncIterator[Team]:
        """
        Get information about teams of an organization.

        https://docs.github.com/en/rest/teams/teams#list-teams

        Args:
            organization: GitHub organization to use

        Raises:
            `httpx.HTTPStatusError` if there was an error in the request
        """
        api = f"/orgs/{organization}/teams"
        params = {
            "per_page": "100",
        }

        async for response in self._client.get_all(api, params=params):
            response.raise_for_status()

            for team in response.json():
                yield Team.from_dict(team)

    async def create(
        self,
        organization: str,
        name: str,
        *,
        description: Optional[str] = None,
        maintainers: Optional[Iterable[str]] = None,
        repo_names: Optional[Iterable[str]] = None,
        privacy: Optional[TeamPrivacy] = None,
        parent_team_id: Optional[str] = None,
    ) -> Team:
        """
        Create a new team in an organization

        https://docs.github.com/en/rest/teams/teams#create-a-team

        Args:
            organization: GitHub organization to use
            name: The name of the new team.
            description: The description of the team.
            maintainers: List GitHub IDs for organization members who will
                become team maintainers.
            repo_names: The full name (e.g., "organization-name/repository-name"
                ) of repositories to add the team to.
            privacy: The level of privacy this team should have. The options
                are:
                For a non-nested team:
                    * secret - only visible to organization owners and members
                        of this team.
                    * closed - visible to all members of this organization.
                Default: secret

                For a parent or child team:
                    * closed - visible to all members of this organization.
                Default for child team: closed
            parent_team_id: The ID of a team to set as the parent team.

        Raises:
            `httpx.HTTPStatusError` if there was an error in the request
        """
        api = f"/orgs/{organization}/teams"
        data = {"name": name}
        if description:
            data["description"] = description
        if maintainers:
            data["maintainers"] = list(maintainers)
        if repo_names:
            data["repo_names"] = list(repo_names)
        if privacy:
            data["privacy"] = privacy.value
        if parent_team_id:
            data["parent_team_id"] = parent_team_id

        response = await self._client.post(api, data=data)
        response.raise_for_status()
        return Team.from_dict(response.json())

    async def get(
        self,
        organization: str,
        team: str,
    ) -> Team:
        """
        Gets a team using the team's slug. GitHub generates the slug from the
        team name.

        https://docs.github.com/en/rest/teams/teams#get-a-team-by-name

        Args:
            organization: GitHub organization to use
            team: The team slug of the team.

        Raises:
            `httpx.HTTPStatusError` if there was an error in the request
        """
        api = f"/orgs/{organization}/teams/{team}"
        response = await self._client.get(api)
        response.raise_for_status()
        return Team.from_dict(response.json())

    async def update(
        self,
        organization: str,
        team: str,
        *,
        name: str,
        description: Optional[str] = None,
        privacy: Optional[TeamPrivacy] = None,
        parent_team_id: Optional[str] = None,
    ) -> Team:
        """
        Update team information in an organization

        https://docs.github.com/en/rest/teams/teams#update-a-team

        Args:
            organization: GitHub organization to use
            team: The slug of the team name.
            name: The name of the team.
            description: The description of the team.
            privacy: The level of privacy this team should have. The options
                are:
                For a non-nested team:
                    * secret - only visible to organization owners and members
                        of this team.
                    * closed - visible to all members of this organization.
                Default: secret

                For a parent or child team:
                    * closed - visible to all members of this organization.
                Default for child team: closed
            parent_team_id: The ID of a team to set as the parent team.

        Raises:
            `httpx.HTTPStatusError` if there was an error in the request
        """
        api = f"/orgs/{organization}/teams/{team}"
        data = {}
        if name:
            data["name"] = name
        if description:
            data["description"] = description
        if privacy:
            data["privacy"] = privacy.value
        if parent_team_id:
            data["parent_team_id"] = parent_team_id

        response = await self._client.post(api, data=data)
        response.raise_for_status()
        return Team.from_dict(response.json())

    async def delete(
        self,
        organization: str,
        team: str,
    ) -> None:
        """
        Delete a new team of an organization

        https://docs.github.com/en/rest/teams/teams#delete-a-team

        Args:
            organization: GitHub organization to use
            team: The slug of the team name.

        Raises:
            `httpx.HTTPStatusError` if there was an error in the request
        """
        api = f"/orgs/{organization}/teams/{team}"
        response = await self._client.delete(api)
        response.raise_for_status()

    async def members(
        self,
        organization: str,
        team: str,
    ) -> AsyncIterator[User]:
        """
        Get all members of a team. Team members will include the members of
        child teams.

        https://docs.github.com/en/rest/teams/members#list-team-members

        Args:
            organization: GitHub organization to use
            team: The slug of the team name.

        Raises:
            `httpx.HTTPStatusError` if there was an error in the request
        """
        api = f"/orgs/{organization}/teams/{team}/members"
        params = {
            "per_page": "100",
        }

        async for response in self._client.get_all(api, params=params):
            response.raise_for_status()

            for member in response.json():
                yield User.from_dict(member)

    async def update_member(
        self,
        organization: str,
        team: str,
        username: str,
        *,
        role: TeamRole = TeamRole.MEMBER,
    ) -> None:
        """
        Add or update a member of a team.

        https://docs.github.com/en/rest/teams/members#add-or-update-team-membership-for-a-user

        Args:
            organization: GitHub organization to use
            team: The slug of the team name.
            username: The handle for the GitHub user account.
            role: The role that this user should have in the team.
                Default: member. Can be one of: member, maintainer.

        Raises:
            `httpx.HTTPStatusError` if there was an error in the request
        """
        api = f"/orgs/{organization}/teams/{team}/memberships/{username}"
        data = {"role": role.value}
        response = await self._client.put(api, data=data)
        response.raise_for_status()

    # add_member is the same API as update_member
    add_member = update_member

    async def remove_member(
        self,
        organization: str,
        team: str,
        username: str,
    ) -> None:
        """
        Remove a member from a team.

        https://docs.github.com/en/rest/teams/members#remove-team-membership-for-a-user

        Args:
            organization: GitHub organization to use
            team: The slug of the team name.
            username: The handle for the GitHub user account.

        Raises:
            `httpx.HTTPStatusError` if there was an error in the request
        """
        api = f"/orgs/{organization}/teams/{team}/memberships/{username}"
        response = await self._client.delete(api)
        response.raise_for_status()

    async def repositories(
        self,
        organization: str,
        team: str,
    ) -> AsyncIterator[Repository]:
        """
        Lists a team's repositories visible to the authenticated user.

        https://docs.github.com/en/rest/teams/teams#list-team-repositories

        Args:
            organization: GitHub organization to use
            team: The slug of the team name.

        Raises:
            `httpx.HTTPStatusError` if there was an error in the request
        """
        api = f"/orgs/{organization}/teams/{team}/repos"
        params = {
            "per_page": "100",
        }

        async for response in self._client.get_all(api, params=params):
            response.raise_for_status()

            for repo in response.json():
                yield Repository.from_dict(repo)

    async def update_permission(
        self,
        organization: str,
        team: str,
        repository: str,
        permission: Permission,
    ) -> None:
        """
        Add or update team repository permissions

        Args:
            organization: GitHub organization to use
            team: The slug of the team name.
            repository: GitHub repository (only name) to add or change
                permissions on.
            permission: The permission to grant the team on the repository.

        Raises:
            `httpx.HTTPStatusError` if there was an error in the request
        """
        api = (
            f"/orgs/{organization}/teams/{team}/repos/{organization}/"
            f"{repository}"
        )
        data = {"permission": permission.value}
        response = await self._client.put(api, data=data)
        response.raise_for_status()

    add_permission = update_permission



# Copyright (C) 2020-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# pylint: disable=no-name-in-module,no-member

import builtins
import sys
import unittest
from typing import Any, AsyncIterator, Awaitable, Iterable
from unittest import TestCase
from unittest.mock import MagicMock

if sys.version_info.minor < 10:
    # aiter and anext have been added in Python 3.10
    def aiter(obj):  # pylint: disable=redefined-builtin
        return obj.__aiter__()

    def anext(obj):  # pylint: disable=redefined-builtin
        return obj.__anext__()

else:

    def aiter(obj):  # pylint: disable=redefined-builtin
        return builtins.aiter(obj)

    def anext(obj):  # pylint: disable=redefined-builtin
        return builtins.anext(obj)


if sys.version_info.minor > 7:
    from unittest import IsolatedAsyncioTestCase
    from unittest.mock import AsyncMock
else:

    @unittest.skip("Async Tests not available for Python 3.7")
    class IsolatedAsyncioTestCase(TestCase):
        pass

    class AsyncMock(MagicMock):
        # implement all necessary methods to keep Pylint happy

        def __init__(self, *args: Any, **kw: Any) -> None:
            super().__init__(*args, **kw)

            self.__aenter__ = self
            self.__aexit__ = self

        def assert_awaited(self):
            pass

        def assert_awaited_once(self):
            pass

        def assert_awaited_with(self, *args, **kwargs):
            pass

        def assert_awaited_once_with(self, *args, **kwargs):
            pass

        def assert_any_await(self, *args, **kwargs):
            pass

        def assert_has_awaits(self, calls, any_order=False):
            pass

        def assert_not_awaited(self):
            pass


class AsyncIteratorMock(AsyncIterator):
    def __init__(self, iterable: Iterable[Any]) -> None:
        self.iterator = iter(iterable)

    async def __anext__(self) -> Awaitable[Any]:
        try:
            return next(self.iterator)
        except StopIteration:
            raise StopAsyncIteration() from None


__all__ = ["IsolatedAsyncioTestCase", "AsyncMock"]



# Copyright (C) 2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
from pathlib import Path
from typing import (
    AsyncContextManager,
    AsyncIterator,
    ContextManager,
    Iterable,
    Iterator,
    Optional,
    Tuple,
    Union,
)

import httpx

from pontos.github.api.client import GitHubAsyncREST
from pontos.github.api.helper import JSON_OBJECT
from pontos.github.models.release import Release
from pontos.helper import (
    AsyncDownloadProgressIterable,
    DownloadProgressIterable,
    download,
    download_async,
    upload,
)


class GitHubAsyncRESTReleases(GitHubAsyncREST):
    async def create(
        self,
        repo: str,
        tag: str,
        *,
        body: Optional[str] = None,
        name: Optional[str] = None,
        target_commitish: Optional[str] = None,
        draft: Optional[bool] = False,
        prerelease: Optional[bool] = False,
    ) -> Release:
        """
        Create a new GitHub release

        https://docs.github.com/en/rest/releases/releases#create-a-release

        Args:
            repo: GitHub repository (owner/name) to use
            tag: The git tag for the release
            body: Content of the changelog for the release
            name: name of the release, e.g. 'pontos 1.0.0'
            target_commitish: Only needed when tag is not there yet
            draft: If the release is a draft. False by default.
            prerelease: If the release is a pre release. False by default.

        Raises:
            httpx.HTTPStatusError if the request was invalid
        """
        data = {
            "tag_name": tag,
            "draft": draft,
            "prerelease": prerelease,
        }
        if name is not None:
            data["name"] = name
        if body is not None:
            data["body"] = body
        if target_commitish is not None:
            data["target_commitish"] = target_commitish

        api = f"/repos/{repo}/releases"
        response = await self._client.post(api, data=data)
        response.raise_for_status()
        return Release.from_dict(response.json())

    async def exists(self, repo: str, tag: str) -> bool:
        """
        Check wether a GitHub release exists by tag

        Args:
            repo: GitHub repository (owner/name) to use
            tag: The git tag for the release

        Returns:
            True if the release exists
        """
        api = f"/repos/{repo}/releases/tags/{tag}"
        response = await self._client.get(api)
        return response.is_success

    async def get(self, repo: str, tag: str) -> Release:
        """
        Get data of a GitHub release by tag

        https://docs.github.com/en/rest/releases/releases#get-a-release-by-tag-name

        Args:
            repo: GitHub repository (owner/name) to use
            tag: The git tag for the release

        Raises:
            httpx.HTTPStatusError if the request was invalid
        """
        api = f"/repos/{repo}/releases/tags/{tag}"
        response = await self._client.get(api)
        response.raise_for_status()
        return Release.from_dict(response.json())

    def download_release_tarball(
        self, repo: str, tag: str
    ) -> AsyncContextManager[AsyncDownloadProgressIterable[bytes]]:
        """
        Download a release tarball (tar.gz) file

        Args:
            repo: GitHub repository (owner/name) to use
            tag: The git tag for the release

        Raises:
            HTTPStatusError if the request was invalid
        """
        api = f"https://github.com/{repo}/archive/refs/tags/{tag}.tar.gz"
        return download_async(self._client.stream(api))

    def download_release_zip(
        self,
        repo: str,
        tag: str,
    ) -> AsyncContextManager[AsyncDownloadProgressIterable[bytes]]:
        """
        Download a release zip file

        Args:
            repo: GitHub repository (owner/name) to use
            tag: The git tag for the release

        Raises:
            HTTPStatusError if the request was invalid
        """
        api = f"https://github.com/{repo}/archive/refs/tags/{tag}.zip"
        return download_async(self._client.stream(api))

    async def download_release_assets(
        self,
        repo: str,
        tag: str,
        *,
        match_pattern: Optional[str] = None,
    ) -> AsyncIterator[
        Tuple[str, AsyncContextManager[AsyncDownloadProgressIterable[bytes]]]
    ]:
        """
        Download release assets

        Args:
            repo: GitHub repository (owner/name) to use
            tag: The git tag for the release
            match_pattern: Optional pattern which the name of the available
                artifact must match. For example "*.zip". Allows to download
                only specific artifacts.

        Raises:
            HTTPError if the request was invalid

        Example:
            .. code-block:: python

            async def download_asset(name, download_cm):
                async with download_cm as iterator:
                    with Path(name).open("wb") as f:
                        async for content, progress in iterator:
                            f.write(content)
                            print(name, progress)


            tasks = []
            async for name, download_cm in api.download_release_assets(
                "foo/bar, "v1.2.3",
            ):
                tasks.append(asyncio.create_task(
                    download_asset(name, download_cm)
                )

            await asyncio.gather(*tasks)

        """
        release = await self.get(repo, tag)
        assets_url = release.assets_url
        if not assets_url:
            raise RuntimeError("assets URL not found")

        response = await self._client.get(assets_url)
        response.raise_for_status()

        assets_json = response.json()
        for asset_json in assets_json:
            asset_url: str = asset_json.get("browser_download_url", "")
            name: str = asset_json.get("name", "")

            if match_pattern and not Path(name).match(match_pattern):
                continue

            yield name, download_async(self._client.stream(asset_url))

    async def upload_release_assets(
        self,
        repo: str,
        tag: str,
        files: Iterable[Union[Path, Tuple[Path, str]]],
    ) -> AsyncIterator[Path]:
        """
        Upload release assets asynchronously

        Args:
            repo: GitHub repository (owner/name) to use
            tag: The git tag for the release
            files: An iterable of file paths or an iterable of tuples
                containing a file path and content types to upload as an asset

        Returns:
            yields each file after its upload is finished

        Raises:
            HTTPError if an upload request was invalid

        Example:
            .. code-block:: python

            files = (Path("foo.txt"), Path("bar.txt"),)
            async for uploaded_file in api.upload_release_assets(
                "foo/bar", "1.2.3", files
            ):
                print(f"Uploaded: {uploaded_file}")

            files = [
                (Path("foo.txt"), "text/ascii"),
                (Path("bar.pdf"), "application/pdf"),
            ]
            async for uploaded_file in api.upload_release_assets(
                "foo/bar", "1.2.3", files
            ):
               print(f"Uploaded: {uploaded_file}")
        """
        release = await self.get(repo, tag)
        asset_url = release.upload_url.replace("{?name,label}", "")

        async def upload_file(
            file_path, content_type
        ) -> Tuple[httpx.Response, Path]:
            response = await self._client.post(
                asset_url,
                params={"name": file_path.name},
                content_type=content_type,
                content=upload(file_path),
            )
            return response, file_path

        tasks = []
        for file_path in files:
            if isinstance(file_path, Tuple):
                file_path, content_type = file_path
            else:
                content_type = "application/octet-stream"

            tasks.append(upload_file(file_path, content_type))

        for coroutine in asyncio.as_completed(tasks):
            response, file_path = await coroutine
            response.raise_for_status()
            yield file_path


class GitHubRESTReleaseMixin:
    def create_tag(
        self,
        repo: str,
        tag: str,
        message: str,
        git_object: str,
        name: str,
        email: str,
        *,
        git_object_type: Optional[str] = None,
        date: Optional[str] = None,
    ) -> str:
        """
        Create a github tag

        Args:
            repo: The name of the repository.
                The name is not case sensitive.
            tag: The tag's name.
                This is typically a version (e.g., "v0.0.1").
            message: The tag message.
            git_object: The SHA of the git object this is tagging.
            git_object_type: The type of the object we're tagging.
                Normally this is a commit
                but it can also be a tree or a blob.
            name: The name of the author of the tag
            email: The email of the author of the tag
            date: When this object was tagged.
                This is a timestamp in
                ISO 8601 format: YYYY-MM-DDTHH:MM:SSZ.

        Return:
            Tag sha
        """

        if not git_object_type:
            git_object_type = "commit"

        data = {
            "repo": repo,
            "tag": tag,
            "message": message,
            "object": git_object,
            "type": git_object_type,
            "tagger": {
                "name": name,
                "email": email,
            },
        }

        if date:
            data["tagger"]["date"] = date

        api = f"/repos/{repo}/git/tags"
        response: httpx.Response = self._request(
            api, data=data, request=httpx.post
        )
        response.raise_for_status()
        return response.json()["object"]["sha"]

    def create_tag_reference(
        self,
        repo: str,
        tag: str,
        sha: str,
    ) -> None:
        """
        Create git tag reference (A real tag in git).

        Args:
            repo: The name of the repository.
                The name is not case sensitive.
            tag: Github tag name.
            sha: The SHA1 value for this Github tag.
        """

        data = {
            "repo": repo,
            "ref": f"refs/tags/{tag}",
            "sha": sha,
        }

        api = f"/repos/{repo}/git/refs"
        response: httpx.Response = self._request(
            api, data=data, request=httpx.post
        )
        response.raise_for_status()

    def create_release(
        self,
        repo: str,
        tag: str,
        *,
        body: Optional[str] = None,
        name: Optional[str] = None,
        target_commitish: Optional[str] = None,
        draft: Optional[bool] = False,
        prerelease: Optional[bool] = False,
    ):
        """
        Args:
            repo: GitHub repository (owner/name) to use
            tag: The git tag for the release
            body: Content of the changelog for the release
            name: name of the release, e.g. 'pontos 1.0.0'
            target_commitish: Only needed when tag is not there yet
            draft: If the release is a draft. False by default.
            prerelease: If the release is a pre release. False by default.
        """
        data = {
            "tag_name": tag,
            "draft": draft,
            "prerelease": prerelease,
        }
        if name is not None:
            data["name"] = name
        if body is not None:
            data["body"] = body
        if target_commitish is not None:
            data["target_commitish"] = target_commitish

        api = f"/repos/{repo}/releases"
        response: httpx.Response = self._request(
            api, data=data, request=httpx.post
        )
        response.raise_for_status()

    def release_exists(self, repo: str, tag: str) -> bool:
        """
        Check wether a GitHub release exists by tag

        Args:
            repo: GitHub repository (owner/name) to use
            tag: The git tag for the release

        Returns:
            True if the release exists
        """
        api = f"/repos/{repo}/releases/tags/{tag}"
        response: httpx.Response = self._request(api)
        return response.is_success

    def release(self, repo: str, tag: str) -> JSON_OBJECT:
        """
        Get data of a GitHub release by tag

        Args:
            repo: GitHub repository (owner/name) to use
            tag: The git tag for the release

        Raises:
            HTTPError if the request was invalid
        """
        api = f"/repos/{repo}/releases/tags/{tag}"
        response: httpx.Response = self._request(api)
        response.raise_for_status()
        return response.json()

    def download_release_tarball(
        self, repo: str, tag: str, destination: Path
    ) -> ContextManager[DownloadProgressIterable]:
        """
        Download a release tarball (tar.gz) file

        Args:
            repo: GitHub repository (owner/name) to use
            tag: The git tag for the release
            destination: A path where to store the downloaded file

        Raises:
            HTTPError if the request was invalid
        """
        api = f"https://github.com/{repo}/archive/refs/tags/{tag}.tar.gz"
        return download(api, destination)

    def download_release_zip(
        self, repo: str, tag: str, destination: Path
    ) -> ContextManager[DownloadProgressIterable]:
        """
        Download a release zip file

        Args:
            repo: GitHub repository (owner/name) to use
            tag: The git tag for the release
            destination: A path where to store the downloaded file

        Raises:
            HTTPError if the request was invalid
        """
        api = f"https://github.com/{repo}/archive/refs/tags/{tag}.zip"
        return download(api, destination)

    def download_release_assets(
        self,
        repo: str,
        tag: str,
    ) -> Iterator[DownloadProgressIterable]:
        """
        Download release assets

        Args:
            repo: GitHub repository (owner/name) to use
            tag: The git tag for the release

        Raises:
            HTTPError if the request was invalid
        """
        release_json = self.release(repo, tag)
        assets_url = release_json.get("assets_url")
        if not assets_url:
            return

        response: httpx.Response = self._request_internal(assets_url)
        response.raise_for_status()
        assets_json = response.json()
        for asset_json in assets_json:
            asset_url: str = asset_json.get("browser_download_url", "")
            name: str = asset_json.get("name", "")
            if name.endswith(".zip") or name.endswith(".tar.gz"):
                with download(
                    asset_url,
                    Path(name),
                ) as progress:
                    yield progress

    def upload_release_assets(
        self,
        repo: str,
        tag: str,
        files: Iterable[Union[Path, Tuple[Path, str]]],
    ) -> Iterator[Path]:
        # pylint: disable=line-too-long
        """
        Upload release assets one after another

        Args:
            repo: GitHub repository (owner/name) to use
            tag: The git tag for the release
            files: An iterable of file paths or an iterable of tuples
                containing a file path and content types to upload as an asset

        Returns:
            yields each file after its upload is finished

        Raises:
            HTTPError if an upload request was invalid

        Examples:
            >>> files = (Path("foo.txt"), Path("bar.txt"),)
            >>> for uploaded_file in git.upload_release_assets("foo/bar", "1.2.3", files):
            >>>    print(f"Uploaded: {uploaded_file}")

            >>> files = [(Path("foo.txt"), "text/ascii"), (Path("bar.pdf"), "application/pdf")]
            >>> for uploaded_file in git.upload_release_assets("foo/bar", "1.2.3", files):
            >>>    print(f"Uploaded: {uploaded_file}")
        """
        github_json = self.release(repo, tag)
        asset_url = github_json["upload_url"].replace("{?name,label}", "")

        for file_path in files:
            if isinstance(file_path, Tuple):
                file_path, content_type = file_path
            else:
                content_type = "application/octet-stream"

            to_upload = file_path.read_bytes()

            response: httpx.Response = self._request_internal(
                asset_url,
                params={"name": file_path.name},
                content=to_upload,
                content_type=content_type,
                request=httpx.post,
            )
            response.raise_for_status()
            yield file_path



# -*- coding: utf-8 -*-
# Copyright (C) 2021-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import argparse
import re

from packaging.version import InvalidVersion, Version


class VersionError(Exception):
    """
    Some error has occurred during version handling
    """


def strip_version(version: str) -> str:
    """
    Strips a leading 'v' from a version string

    E.g. v1.2.3 will be converted to 1.2.3
    """
    if version and version[0] == "v":
        return version[1:]

    return version


def check_develop(version: str) -> str:
    """
    Checks if the given Version is a develop version

    Returns True if yes, False if not
    """
    return True if Version(version).dev is not None else False


def is_version_pep440_compliant(version: str) -> bool:
    """
    Checks if the provided version is a PEP 440 compliant version string
    """
    return version == safe_version(version)


def initialize_default_parser() -> argparse.ArgumentParser:
    """
    Returns a default argument parser containing:
    - verify
    - show
    - update
    """
    parser = argparse.ArgumentParser(
        description="Version handling utilities.",
        prog="version",
    )
    parser.add_argument(
        "--quiet", help="don't print messages", action="store_true"
    )

    subparsers = parser.add_subparsers(
        title="subcommands",
        description="valid subcommands",
        help="additional help",
        dest="command",
    )

    verify_parser = subparsers.add_parser("verify")
    verify_parser.add_argument("version", help="version string to compare")
    subparsers.add_parser("show")

    update_parser = subparsers.add_parser("update")
    update_parser.add_argument("version", help="version string to use")
    update_parser.add_argument(
        "--force",
        help="don't check if version is already set",
        action="store_true",
    )
    update_parser.add_argument(
        "--develop",
        help="indicates if it is a develop version",
        action="store_true",
    )
    return parser


def safe_version(version: str) -> str:
    """
    Returns the version as a string in `PEP440`_ compliant
    format.

    .. _PEP440:
       https://www.python.org/dev/peps/pep-0440
    """
    try:
        return str(Version(version))
    except InvalidVersion:
        version = version.replace(" ", ".")
        return re.sub("[^A-Za-z0-9.]+", "-", version)


def versions_equal(new_version: str, old_version: str) -> bool:
    """
    Checks if new_version and old_version are equal
    """
    return safe_version(old_version) == safe_version(new_version)



# Copyright (C) 2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import contextlib
import io
import json
import unittest
from pathlib import Path
from unittest.mock import MagicMock, patch

from pontos.version.helper import VersionError
from pontos.version.javascript import JavaScriptVersionCommand
from tests.version import use_cwd


class JavaScriptVersionCommandTestCase(unittest.TestCase):
    def test_missing_pyproject_toml_file(self):
        with use_cwd(Path(__file__).parent), self.assertRaisesRegex(
            VersionError, r"^.* not found\.$"
        ):
            JavaScriptVersionCommand()

    @patch("pontos.version.javascript.json", spec=json)
    def test_missing_tool_pontos_version_section(self, json_mock):
        project_file_path = MagicMock(spec=Path).return_value
        project_file_path.exists.return_value = True
        json_mock.load.return_value = json.loads('{"name":"foo", "bar":"bar"}')

        with self.assertRaisesRegex(
            VersionError, r"Version field missing in .*\.$"
        ):
            JavaScriptVersionCommand(project_file_path=project_file_path)

    @patch("pontos.version.javascript.json", spec=json)
    def test_get_version(self, json_mock):
        project_file_path = MagicMock(spec=Path).return_value
        project_file_path.exists.return_value = True
        json_mock.load.return_value = json.loads(
            '{"name":"foo", "version":"1.2.3"}'
        )

        cmd = JavaScriptVersionCommand(project_file_path=project_file_path)
        # pylint: disable=protected-access
        version = cmd.get_current_version()

        self.assertEqual(version, "1.2.3")


class UpdateJavaScriptVersionTestCase(unittest.TestCase):
    def test_update_version_file(self):
        package_file = Path(__file__).parent / "fake_package.json"
        package_file.write_text(
            '{"name":"foo", "version":"1.2.3"}', encoding="utf-8"
        )
        with contextlib.redirect_stdout(io.StringIO()):
            cmd = JavaScriptVersionCommand(project_file_path=package_file)
            cmd.update_version("22.4.0.dev1")

        with package_file.open(mode="r", encoding="utf-8") as fp:
            fake_package = json.load(fp)

        self.assertEqual(fake_package["version"], "22.4.0.dev1")

        package_file.unlink()



# Copyright (C) 2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from argparse import ArgumentParser, Namespace

from pontos.github.api import GitHubAsyncRESTApi
from pontos.github.api.branch import update_from_applied_settings


def add_script_arguments(parser: ArgumentParser) -> None:
    parser.add_argument("repo")
    parser.add_argument("branch")


async def github_script(api: GitHubAsyncRESTApi, args: Namespace) -> int:
    # draft script for updating the branch protections
    branch_protection = await api.branches.protection_rules(
        args.repo, args.branch
    )
    # switch required signatures enabled/disabled
    kwargs = update_from_applied_settings(
        branch_protection,
        # pylint: disable=line-too-long
        required_signatures=not branch_protection.required_signatures.enabled,
    )
    await api.branches.update_protection_rules(
        args.repo,
        args.branch,
        **kwargs,
    )
    return 0



# Copyright (C) 2020-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys
from pathlib import Path
from typing import Tuple

from .__version__ import __version__
from .cmake import CMakeVersionCommand
from .go import GoVersionCommand
from .javascript import JavaScriptVersionCommand
from .python import PythonVersionCommand


def main(leave=True, args=None) -> Tuple[bool, str]:
    available_cmds = [
        ("CMakeLists.txt", CMakeVersionCommand),
        ("pyproject.toml", PythonVersionCommand),
        ("package.json", JavaScriptVersionCommand),
        ("go.mod", GoVersionCommand),
    ]
    for file_name, cmd in available_cmds:
        project_definition_path = Path.cwd() / file_name
        if project_definition_path.exists():
            result = cmd().run(args)
            if leave:
                sys.exit(result)
            return result == 0, file_name
    if leave:
        sys.exit("No command found")
    return False, ""



# -*- coding: utf-8 -*-
# Copyright (C) 2021-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from pontos.terminal.terminal import ConsoleTerminal


def main() -> None:
    term = ConsoleTerminal()

    term.print()
    term.bold_info("pontos - Greenbone Python Utilities and Tools")
    term.print()
    term.print("The following commands are currently available:")
    with term.indent():
        term.bold_info(
            "pontos-release - Release handling "
            "utility for C and Python Projects"
        )
        term.print("usage:")
        with term.indent():
            term.print("pontos-release {prepare,release,sign} -h")
        term.bold_info(
            "pontos-version - Version handling utility "
            "for C, Go and Python Projects"
        )
        term.print("usage:")
        with term.indent():
            term.print("pontos-version {verify,show,update} -h")
        term.bold_info(
            "pontos-update-header - Handling Copyright header "
            "for various file types and licenses"
        )
        term.print("usage:")
        with term.indent():
            term.print("pontos-update-header -h")
        term.bold_info(
            "pontos-changelog - Parse conventional commits in the "
            "current branch, creating CHANGELOG.md file"
        )
        term.print("usage:")
        with term.indent():
            term.print("pontos-changelog -h")
        term.bold_info(
            "pontos-github - Handling GitHub operations, like "
            "Pull Requests (beta)"
        )
        term.print("usage:")
        with term.indent():
            term.print("pontos-github {pr} -h")
        term.print()
        term.info(
            "pontos also comes with a Terminal interface "
            "printing prettier outputs"
        )
        with term.indent():
            term.print('Accessible with import "pontos.terminal"')
        term.info("pontos also comes with git and GitHub APIs")
        with term.indent():
            term.print(
                'Accessible with "import pontos.git" '
                'and "import pontos.github"'
            )

    term.print()
    term.warning(
        'Use the listed commands "help" for more information '
        "and arguments description."
    )


if __name__ == "__main__":
    main()



# Copyright (C) 2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest
from unittest.mock import patch

from pontos.github.api.helper import DEFAULT_TIMEOUT
from pontos.github.script.parser import create_parser


class CreateParserTestCase(unittest.TestCase):
    def test_create_parser(self):
        parser = create_parser()
        args = parser.parse_args(
            ["--token", "123", "--timeout", "456", "my_script"]
        )

        self.assertEqual(args.token, "123")
        self.assertEqual(args.timeout, 456)
        self.assertEqual(args.script, "my_script")

    @patch.dict("os.environ", {"GITHUB_TOKEN": "987"})
    def test_parse_token(self):
        parser = create_parser()
        args = parser.parse_args(["my_script"])
        self.assertEqual(args.token, "987")
        self.assertEqual(args.script, "my_script")

        args = parser.parse_args(["--token", "123", "my_script"])

        self.assertEqual(args.token, "123")
        self.assertEqual(args.script, "my_script")

    def test_parse_timeout(self):
        parser = create_parser()
        args = parser.parse_args(["my_script"])
        self.assertEqual(args.script, "my_script")
        self.assertEqual(args.timeout, DEFAULT_TIMEOUT)

        args = parser.parse_args(["--timeout", "666", "my_script"])

        self.assertEqual(args.timeout, 666)
        self.assertEqual(args.script, "my_script")
